import collections
import math
import pathlib
import pytest
import tempfile

import demes

cwd = pathlib.Path(__file__).parent.resolve()
example_dir = cwd / ".." / "examples"


class TestFromMs:
    def N0(self, *, theta, mu, length):
        return theta / (4 * mu * length)

    def test_ignored_options_have_no_effect(self):
        def check(command, N0=1):
            # with pytest.warns(UserWarning, match="Ignoring unknown args"):
            graph = demes.from_ms(command, N0=N0)
            assert len(graph.pulses) == 0
            assert len(graph.migrations) == 0
            assert len(graph.demes) == 1
            assert len(graph["deme1"].epochs) == 1
            epoch = graph["deme1"].epochs[0]
            assert math.isclose(epoch.start_size, N0)
            assert math.isclose(epoch.end_size, N0)

        for theta in (1, 100.0, 5e6):
            cmd = f"ms 2 1 -t {theta}"
            check(cmd)

        for seeds in ([1, 2, 3], [123, 321, 678]):
            cmd = f"ms 2 1 -t 1.0 -seeds {seeds[0]} {seeds[1]} {seeds[2]}"
            check(cmd)

        for segsites in (0, 1, 1000):
            cmd = f"ms 2 1 -s {segsites}"
            check(cmd)

        for flag in ("-T", "-L"):
            cmd = f"ms 2 1 {flag}"
            check(cmd)

        for precision in (2, 4, 10):
            cmd = f"ms 2 1 -t 1.0 -p {precision}"
            check(cmd)

        for rho, nsites in ([0, 1000], [2.0, 1000], [50.0, 250e6]):
            cmd = f"ms 2 1 -t 1.0 -r {rho} {nsites}"
            check(cmd)

            for f, lambda_ in ([0, 50], [0.5, 200], [1e-5, 3500]):
                cmd = f"ms 2 1 -t 1.0 -r {rho} {nsites} -c {f} {lambda_}"
                check(cmd)

    def test_empty_command(self):
        # One-deme constant-size model.
        graph = demes.from_ms("", N0=1)
        assert len(graph.pulses) == 0
        assert len(graph.migrations) == 0
        assert len(graph.demes) == 1
        assert len(graph["deme1"].epochs) == 1
        epoch = graph["deme1"].epochs[0]
        assert epoch.start_size == epoch.end_size

    def test_structure(self):
        # -I npop n1 n2 ... [4*N0*m]
        def check(command, num_demes: int, migration: bool, N0=1):
            graph = demes.from_ms(command, N0=N0)
            assert len(graph.pulses) == 0
            if migration:
                assert len(graph.migrations) == num_demes * (num_demes - 1)
            else:
                assert len(graph.migrations) == 0
            assert len(graph.demes) == num_demes
            return graph

        N0 = 100
        for npop in (2, 3, 100):
            sample_config = " 0" * (npop - 1)
            cmd = f"ms 2 1 -I {npop} 2 {sample_config}"
            check(cmd, npop, migration=False, N0=N0)
            for M in (1e-5, 1.0, 400):
                assert M <= 4 * N0, "mutation rate too high, use a larger N0"
                cmd = f"ms 2 1 -I {npop} 2 {sample_config} {M}"
                graph = check(cmd, npop, migration=True, N0=N0)
                # The migration rate used with -I is a bit different to
                # the other migration options. The rate provided corresponds
                # to the sum total of all migrations entering any given deme.
                ingress_rate = collections.defaultdict(lambda: 0)
                rate = M / (4 * N0)
                for migration in graph.migrations:
                    assert math.isclose(migration.rate, rate / (npop - 1))
                    ingress_rate[migration.dest] += migration.rate
                for deme, rate_in in ingress_rate.items():
                    assert math.isclose(rate_in, rate), (
                        f"sum of migrations entering {deme} is {rate_in}, "
                        f"expected {rate}"
                    )

        for bad_npop in (0, -1, 1.1, "x", math.inf):
            cmd = f"ms 2 1 -I {bad_npop} 2 2"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_M in (-1, "x", math.inf):
            cmd = f"ms 2 1 -I 2 1 1 {bad_M}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_sample_configuration in ("50 0 0 2", "2 0 0 0 0 0 2"):
            cmd = "ms 2 1 -t 1.0 -I {bad_sample_configuration}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

    def test_growth_rates(self):
        # -G α
        # -eG t α
        # -g i α
        # -eg t i α
        def check(command, num_demes=1, N0=1):
            graph = demes.from_ms(command, N0=N0)
            assert len(graph.pulses) == 0
            assert len(graph.migrations) == 0
            assert len(graph.demes) == num_demes
            return graph

        graph = check("ms 2 1 -t 1.0 -G 0")
        assert len(graph["deme1"].epochs) == 1
        epoch = graph["deme1"].epochs[0]
        assert epoch.start_size == epoch.end_size == 1

        # To convert model that includes growth rates, there must be two or
        # more epochs, otherwise the growth rate applies up to time=inf,
        # and we can't calculate a start_size for the deme.
        for alpha in (-0.01, 1e-5, 0.01, 5):
            cmd = f"ms 2 1 -t 1.0 -G {alpha}"
            with pytest.raises(ValueError, match="growth rate"):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -eG 0 {alpha}"
            with pytest.raises(ValueError, match="growth rate"):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 2 1 1 -eg 0 2 {alpha}"
            with pytest.raises(ValueError, match="growth rate"):
                demes.from_ms(cmd, N0=1)

        def check_2epoch(cmd, T, N1, num_demes=1, name="deme1", N0=1):
            """
            Check the growth rate was applied in the most recent epoch.
            """
            graph = check(cmd, num_demes=num_demes, N0=N0)
            assert len(graph[name].epochs) == 2
            e1, e2 = graph[name].epochs
            assert e1.start_size == e1.end_size == e2.start_size
            assert math.isclose(e2.start_size, N1)
            assert e2.end_size == N0
            for deme in graph.demes:
                if deme.name == name:
                    continue
                assert len(deme.epochs) == 1
                e1 = deme.epochs[0]
                assert e1.start_size == e1.end_size == N0

        T1 = 10
        T2 = 20
        N0 = 100

        for alpha in (-0.01, 1e-5, 0.01, 5):
            N1 = N0 * math.exp(-alpha * T1)
            cmd = f"ms 2 1 -t 1.0 -G {alpha} -eG {T1} 0"
            check_2epoch(cmd, T1, N1, N0=N0)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -g 1 {alpha} -eG {T1} 0"
            check_2epoch(cmd, T1, N1, N0=N0)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -g 1 {alpha} -eg {T1} 1 0"
            check_2epoch(cmd, T1, N1, N0=N0)

            # Change alpha in deme2
            cmd = f"ms 2 1 -t 1.0 -I 2 1 1 -g 2 {alpha} -eG {T1} 0"
            check_2epoch(cmd, T1, N1, num_demes=2, name="deme2", N0=N0)
            cmd = f"ms 2 1 -t 1.0 -I 2 1 1 -g 2 {alpha} -eg {T1} 2 0"
            check_2epoch(cmd, T1, N1, num_demes=2, name="deme2", N0=N0)

        def check_3epoch(cmd, T1, T2, N1, num_demes=1, name="deme1", N0=1):
            """
            Check the growth rate was applied in the middle epoch.
            """
            graph = check(cmd, num_demes=num_demes, N0=N0)
            assert len(graph[name].epochs) == 3
            e1, e2, e3 = graph[name].epochs
            assert e1.start_size == e1.end_size == e2.start_size
            assert math.isclose(e2.start_size, N1)
            assert e2.end_size == N0
            assert e3.start_size == e3.end_size == N0
            for deme in graph.demes:
                if deme.name == name:
                    continue
                assert len(deme.epochs) == 1
                e1 = deme.epochs[0]
                assert e1.start_size == e1.end_size == N0

        for alpha in (-0.01, 1e-5, 0.01, 5):
            N1 = N0 * math.exp(-alpha * (T2 - T1))
            cmd = f"ms 2 1 -t 1.0 -eG {T1} {alpha} -eG {T2} 0"
            check_3epoch(cmd, T1, T2, N1, N0=N0)
            cmd = f"ms 2 1 -t 1.0 -eg {T1} 1 {alpha} -eG {T2} 0"
            check_3epoch(cmd, T1, T2, N1, N0=N0)
            cmd = f"ms 2 1 -t 1.0 -eG {T1} {alpha} -eg {T2} 1 0"
            check_3epoch(cmd, T1, T2, N1, N0=N0)
            cmd = f"ms 2 1 -t 1.0 -eg {T1} 1 {alpha} -eg {T2} 1 0"
            check_3epoch(cmd, T1, T2, N1, N0=N0)

            # Change alpha in deme2
            cmd = f"ms 2 1 -t 1.0 -I 2 1 1 -eg {T1} 2 {alpha} -eg {T2} 2 0"
            check_3epoch(cmd, T1, T2, N1, num_demes=2, name="deme2", N0=N0)
            cmd = f"ms 2 1 -t 1.0 -I 2 1 1 -eg {T1} 2 {alpha} -eG {T2} 0"
            check_3epoch(cmd, T1, T2, N1, num_demes=2, name="deme2", N0=N0)

        for bad_alpha in ("x", math.inf):
            cmd = f"ms 2 1 -t 1.0 -G {bad_alpha}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -G {bad_alpha} -eG 10 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -g 1 {bad_alpha} -eG 10 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -eG 10 {bad_alpha} -eG 20 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -eg 10 1 {bad_alpha} -eG 20 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_time in ("x", math.inf, -10, -0.1):
            cmd = f"ms 2 1 -t 1.0 -eG {bad_time} 0.01 -eG 10 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -eg {bad_time} 1 0.01 -eG 20 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_popid in ("x", math.inf, 0.5, 1000):
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -eg 0 {bad_popid} 0.01 -eG 20 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -g {bad_popid} 0.01 -eG 10 0"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

    def test_size_changes(self):
        # -n, -en, -eN
        # -eN t x
        # -n i x
        # -en t i x
        def check(command, sizes, N0=1):
            graph = demes.from_ms(command, N0=N0)
            assert len(graph.pulses) == 0
            assert len(graph.migrations) == 0
            assert len(graph.demes) == len(sizes)
            for deme, size in zip(graph.demes, sizes):
                assert len(deme.epochs) == 1
                epoch = deme.epochs[0]
                assert math.isclose(epoch.start_size, size)
                assert math.isclose(epoch.end_size, size)

        for size in (0.1, 2, 10.0, 100):
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -n 1 {size}"
            check(cmd, sizes=[size])
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -n 2 {size}"
            check(cmd, sizes=[1, size, 1])
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -n 2 {size} -n 3 {2 * size}"
            check(cmd, sizes=[1, size, 2 * size])
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -en 0 1 {size}"
            check(cmd, sizes=[size])
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -eN 0 {size}"
            check(cmd, sizes=[size, size, size])

        def check_2epoch(command, T, sizes, N0=1):
            graph = demes.from_ms(command, N0=N0)
            assert len(graph.pulses) == 0
            assert len(graph.migrations) == 0
            assert len(graph.demes) == len(sizes)
            for deme, (N1, N2) in zip(graph.demes, sizes):
                assert len(deme.epochs) == 2
                e1, e2 = deme.epochs
                assert math.isclose(e1.start_size, N1)
                assert math.isclose(e1.end_size, N1)
                assert math.isclose(e1.end_time, T * 4 * N0)
                assert math.isclose(e2.start_size, N2)
                assert math.isclose(e2.end_size, N2)
                assert e2.end_time == 0

        # two epoch
        T = 10
        for size in (0.1, 2, 10.0, 100):
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -en {T} 1 {size}"
            check_2epoch(cmd, T, sizes=[(size, 1)])
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -en 0 1 {size} -en {T} 1 1"
            check_2epoch(cmd, T, sizes=[(1, size)])
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -eN {T} {size}"
            check_2epoch(cmd, T, sizes=[(size, 1), (size, 1), (size, 1)])
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -eN 0 {size} -eN {T} 1"
            check_2epoch(cmd, T, sizes=[(1, size), (1, size), (1, size)])

        for bad_size in (-1, "x", math.inf):
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -n 1 {bad_size}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 3 2 0 0 -n 2 {bad_size}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)
            cmd = f"ms 2 1 -t 1.0 -I 1 2 -en 10 1 {bad_size}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

    def test_migration(self):
        # -eM t x
        # -m i j rate
        # -em t i j rate
        # -ma M11 M12 M12 ... M21 ...
        # -ema t npop M11 M12 M12 ... M21 ...
        N0 = 100
        # rate=0 is permissible, but shouldn't result in migration
        for cmd in (
            "-I 2 0 2 -m 1 2 0",
            "-I 2 0 2 -em 0 1 2 0",
            "-I 2 0 2 -em 10 1 2 0",
            "-I 2 0 2 -ma 0 0 0 0",
            "-I 2 0 2 -ma x 0 0 x",
            "-I 2 0 2 -ema 0 2 0 0 0 0",
            "-I 2 0 2 -ema 0 2 x 0 0 x",
            "-I 2 0 2 -ema 10 2 0 0 0 0",
            "-I 2 0 2 -ema 10 2 x 0 0 x",
        ):
            graph = demes.from_ms(cmd, N0=N0)
            assert len(graph.migrations) == 0

        for rate in (0.1, 1e-5):
            # one migration
            for cmd in (
                f"-I 2 0 2 -m 1 2 {rate}",
                f"-I 2 0 2 -em 0 1 2 {rate}",
                f"-I 2 0 2 -em 10 1 2 {rate}",
                f"-I 2 0 2 -ma 0 {rate} 0 0",
                f"-I 2 0 2 -ma x {rate} 0 x",
                f"-I 2 0 2 -ema 0 2 0 {rate} 0 0",
                f"-I 2 0 2 -ema 0 2 x {rate} 0 x",
                f"-I 2 0 2 -ema 10 2 0 {rate} 0 0",
                f"-I 2 0 2 -ema 10 2 x {rate} 0 x",
            ):
                graph = demes.from_ms(cmd, N0=N0)
                assert len(graph.migrations) == 1
                assert math.isclose(graph.migrations[0].rate, rate / (4 * N0))

            # two migrations
            for cmd in (
                f"-I 2 0 2 -m 1 2 {rate} -m 2 1 {rate}",
                f"-I 2 0 2 -em 0 1 2 {rate} -em 0 2 1 {rate}",
                f"-I 2 0 2 -em 10 1 2 {rate} -em 10 2 1 {rate}",
                f"-I 2 0 2 -ma 0 {rate} {rate} 0",
                f"-I 2 0 2 -ma x {rate} {rate} x",
                f"-I 2 0 2 -ema 0 2 0 {rate} {rate} 0",
                f"-I 2 0 2 -ema 0 2 x {rate} {rate} x",
                f"-I 2 0 2 -ema 10 2 0 {rate} {rate} 0",
                f"-I 2 0 2 -ema 10 2 x {rate} {rate} x",
            ):
                graph = demes.from_ms(cmd, N0=N0)
                assert len(graph.migrations) == 2
                for migration in graph.migrations:
                    assert math.isclose(migration.rate, rate / (4 * N0))

            # two migrations at distinct times
            T1 = 10
            T2 = 20
            for on1 in (
                # turn migration on at t=0
                f"-m 1 2 {rate}",
                f"-ma 0 {rate} 0 0",
                f"-em 0 1 2 {rate}",
                f"-ema 0 2 0 {rate} 0 0",
            ):
                for off in (
                    # turn migration off at t=T1
                    f"-em {T1} 1 2 0",
                    f"-ema {T1} 2 0 0 0 0",
                ):
                    for on2 in (
                        # turn migration on again, with twice the rate, at t=T2
                        f"-em {T2} 1 2 {rate*2}",
                        f"-ema {T2} 2 0 {rate*2} 0 0",
                    ):

                        cmd = f"-I 2 0 2 {on1} {off} {on2}"
                        graph = demes.from_ms(cmd, N0=N0)
                        assert len(graph.migrations) == 2
                        # Check the migration details, but avoid depending on the
                        # ordering of the migrations list.
                        m1, m2 = sorted(graph.migrations, key=lambda m: m.end_time)
                        assert m1.end_time == 0
                        assert math.isclose(T1 * 4 * N0, m1.start_time)
                        assert math.isclose(T2 * 4 * N0, m2.end_time)
                        assert math.isinf(m2.start_time)
                        assert math.isclose(m1.rate, rate / (4 * N0))
                        assert math.isclose(m2.rate, 2 * rate / (4 * N0))

        # number of populations doesn't match migration matrix size
        for bad_npop in (1, 20):
            for cmd in (
                f"-I {bad_npop} {' 0'*(bad_npop-1)} 2 -ma 0 0 0 0",
                f"-I {bad_npop} {' 0'*(bad_npop-1)} 2 -ema 0 2 0 0 0 0",
                f"-I {bad_npop} {' 0'*(bad_npop-1)} 2 -ema 10 2 0 0 0 0",
                f"-I 2 2 -ema 0 {bad_npop} {' 0'*bad_npop}",
                f"-I 2 2 -ema 10 {bad_npop} {' 0'*bad_npop}",
            ):
                with pytest.raises(ValueError):
                    demes.from_ms(cmd, N0=1)

        # invalid npop value
        for bad_npop in (-1, 1.1, "x", math.inf):
            for cmd in (
                f"-I 2 2 -ema 0 {bad_npop} 0 0 0 0",
                f"-I 2 2 -ema 10 {bad_npop} 0 0 0 0",
            ):
                with pytest.raises(ValueError):
                    demes.from_ms(cmd, N0=1)

        # invalid population ids
        for bad_popid in (-1, 1.1, "x", math.inf):
            for cmd in (
                f"-I 2 0 2 -m 2 {bad_popid} 0.01",
                f"-I 2 0 2 -m {bad_popid} 2 0.01",
                f"-I 2 0 2 -m {bad_popid} {bad_popid} 0.01",
                f"-I 2 0 2 -em 0 2 {bad_popid} 0.01",
                f"-I 2 0 2 -em 0 {bad_popid} 2 0.01",
                f"-I 2 0 2 -em 0 {bad_popid} {bad_popid} 0.01",
                f"-I 2 0 2 -em 10 2 {bad_popid} 0.01",
                f"-I 2 0 2 -em 10 {bad_popid} 2 0.01",
                f"-I 2 0 2 -em 10 {bad_popid} {bad_popid} 0.01",
                f"-I 10 0 0 0 0 0 0 0 0 2 0 0.4 -m {bad_popid} 7 0.01",
                f"-I 10 0 0 0 0 0 0 0 2 0 0 0.5 -em 0 7 {bad_popid} 0.01",
                f"-I 10 0 0 0 0 0 0 2 0 0 0 0.6 -em 10 7 {bad_popid} 0.01",
            ):
                with pytest.raises(ValueError):
                    demes.from_ms(cmd, N0=1)

        # bad population ids
        for cmd in (
            # ids too high for number of populations
            "-I 2 0 2 -m 2 3 0.01",
            "-I 2 0 2 -m 3 2 0.01",
            "-I 2 0 2 -m 3 4 0.01",
            "-I 2 0 2 -em 0 2 3 0.01",
            "-I 2 0 2 -em 0 3 2 0.01",
            "-I 2 0 2 -em 0 3 4 0.01",
            "-I 2 0 2 -em 10 2 3 0.01",
            "-I 2 0 2 -em 10 3 2 0.01",
            "-I 2 0 2 -em 10 3 4 0.01",
            "-I 10 0 0 0 0 0 0 0 0 2 0 0.4 -m 20 7 0.01",
            "-I 10 0 0 0 0 0 0 0 2 0 0 0.5 -em 0 7 20 0.01",
            "-I 10 0 0 0 0 0 0 2 0 0 0 0.6 -em 10 7 20 0.01",
            # duplicate ids
            "-I 2 0 0 -m 2 2 0.01",
            "-I 2 0 0 -em 0 2 2 0.01",
            "-I 2 0 0 -em 10 2 2 0.01",
        ):
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_time in (-1, math.inf):
            for cmd in (
                f"-I 2 2 -ema {bad_time} 0 0 0 0",
                f"-I 2 2 -ema {bad_time} 0 0.1 0 0",
                f"-I 2 2 -em {bad_time} 1 2 0",
                f"-I 2 2 -em {bad_time} 1 2 0.1",
            ):
                with pytest.raises(ValueError):
                    demes.from_ms(cmd, N0=1)

        for bad_rate in (-1, math.inf):
            for cmd in (
                f"-I 2 2 -m 1 2 {bad_rate}",
                f"-I 2 2 -em 0 1 2 {bad_rate}",
                f"-I 2 2 -em 10 1 2 {bad_rate}",
                f"-I 2 0 2 -ma 0 {bad_rate} 0 0",
                f"-I 2 0 2 -ema 0 2 0 {bad_rate} 0 0",
                f"-I 2 0 2 -ema 10 2 0 {bad_rate} 0 0",
            ):
                with pytest.raises(ValueError):
                    demes.from_ms(cmd, N0=1)

    def test_asymmetric_migration(self):
        # In ms, M[i][j] is the (forwards time) fraction of deme i which is
        # made up of migrants from deme j each generation.
        # See ./ms_asymmetric_migration.sh for a test that confirms this
        # using the ms commands below.

        def check(cmd):
            graph = demes.from_ms(cmd, N0=1000)
            assert len(graph.demes) == 3
            assert len(graph.migrations) == 1
            assert graph.migrations[0].source == "deme3"
            assert graph.migrations[0].dest == "deme2"

        # Set M[3][2] = 1.0 using the -em option.
        cmd1 = (
            "ms 2 1000 "
            "-t 1.0 "
            "-I 3 1 0 1 "
            "-em 0.0 3 2 1.0 "
            "-ej 100.0 2 1 "
            "-eM 100.0 0.0"
        )
        check(cmd1)

        # Set M[3][2] = 1.0 using the -ema option
        cmd2 = (
            "ms 2 1000 "
            "-t 1.0 "
            "-I 3 1 0 1 "
            "-ema 0.0 3 x 0 0 0 x 0 0 1.0 x "
            "-ej 100.0 2 1 "
            "-eM 100.0 0.0"
        )
        check(cmd2)

        # Set M[3][2] = 1.0 using the -ma option
        cmd3 = (
            "ms 2 1000 "
            "-t 1.0 "
            "-I 3 1 0 1 "
            "-ma x 0 0 0 x 0 0 1.0 x "
            "-ej 100.0 2 1 "
            "-eM 100.0 0.0"
        )
        check(cmd3)

        # Set M[3][2] = 1.0 using the -m option.
        cmd4 = (
            "ms 2 1000 "
            "-t 1.0 "
            "-I 3 1 0 1 "
            "-m 3 2 1.0 "
            "-ej 100.0 2 1 "
            "-eM 100.0 0.0"
        )
        check(cmd4)

    def test_admixture(self):
        # -es t i p
        # from_ms() creates a new deme (say, j) with end_time=t,
        # and a pulse of migration from deme j into deme i at time t.
        N0 = 100
        for t in (0.01, 1, 10):
            for p in (0.01, 0.5, 0.9):
                for num_demes, dest_id in (
                    (1, 1),
                    (2, 2),
                    (10, 3),
                ):
                    cmd = (
                        f"-I {num_demes} {' 0'*(num_demes - 1)} 2 "
                        "-eN 0 12 "
                        f"-es {t} {dest_id} {p}"
                    )
                    graph = demes.from_ms(cmd, N0=N0)
                    assert len(graph.demes) == num_demes + 1
                    assert len(graph.migrations) == 0
                    assert len(graph.pulses) == 1
                    dest = f"deme{dest_id}"
                    source = f"deme{num_demes + 1}"
                    # pulse properties
                    pulse = graph.pulses[0]
                    assert pulse.source == source
                    assert pulse.dest == dest
                    assert math.isclose(pulse.proportion, 1 - p)
                    assert math.isclose(pulse.time, t * 4 * N0)
                    # source deme properties
                    assert math.isinf(graph[source].start_time)
                    assert math.isclose(graph[source].end_time, t * 4 * N0)
                    assert len(graph[source].epochs) == 1
                    assert math.isclose(graph[source].epochs[0].start_size, N0)
                    assert math.isclose(graph[source].epochs[0].end_size, N0)
                    # other demes
                    for deme in graph.demes:
                        if deme.name == source:
                            continue
                        assert math.isinf(deme.start_time)
                        assert deme.end_time == 0
                        assert len(deme.epochs) == 1
                        assert math.isclose(deme.epochs[0].start_size, 12 * N0)
                        assert math.isclose(deme.epochs[0].end_size, 12 * N0)

        for bad_t in (-1, 0, math.inf):
            cmd = f"-es {bad_t} 1 0.1"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_popid in (-1, 200, math.inf):
            cmd = f"-es 1.0 {bad_popid} 0.1"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_p in (-1, 2, math.inf):
            cmd = f"-es 1.0 1 {bad_p}"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

    def test_split(self):
        # -ej t i j
        # from_ms() turns this into a "branch" event where
        # deme i branches from deme j at time t.
        N0 = 100
        for t in (0.01, 1, 10):
            for num_demes, i, j in (
                (2, 1, 2),
                (2, 2, 1),
                (10, 3, 8),
            ):
                cmd = (
                    f"-I {num_demes} {' 0'*(num_demes - 1)} 2 1.0 "
                    "-eN 0 12 "
                    f"-ej {t} {i} {j}"
                )
                graph = demes.from_ms(cmd, N0=N0)
                assert len(graph.demes) == num_demes
                assert len(graph.migrations) == num_demes * (num_demes - 1)
                assert len(graph.pulses) == 0
                child = f"deme{i}"
                assert math.isclose(graph[child].start_time, t * 4 * N0)
                assert graph[child].end_time == 0
                assert len(graph[child].epochs) == 1
                assert math.isclose(graph[child].epochs[0].start_size, 12 * N0)
                assert math.isclose(graph[child].epochs[0].end_size, 12 * N0)
                for deme in graph.demes:
                    if deme.name == child:
                        continue
                    assert math.isinf(deme.start_time)
                    assert deme.end_time == 0
                    assert len(deme.epochs) == 1
                    assert math.isclose(deme.epochs[0].start_size, 12 * N0)
                    assert math.isclose(deme.epochs[0].end_size, 12 * N0)
                # check migrations
                rate = 1 / (4 * N0)
                for migration in graph.migrations:
                    assert math.isclose(migration.rate, rate / (num_demes - 1))
                    # migrations should match the start/end time of the participants
                    start_time = min(
                        graph[migration.source].start_time,
                        graph[migration.dest].start_time,
                    )
                    end_time = max(
                        graph[migration.source].end_time, graph[migration.dest].end_time
                    )
                    assert start_time == migration.start_time
                    assert end_time == migration.end_time

        # check a model with successive branching
        T1, T2, T3, T4 = 1.0, 2.0, 3.0, 4.0
        cmd = (
            "-I 5 1 1 1 1 1 "
            f"-ej {T1} 1 2 "
            f"-ej {T2} 2 3 "
            f"-ej {T3} 3 4 "
            f"-ej {T4} 4 5 "
        )
        graph = demes.from_ms(cmd, N0=N0)
        assert len(graph.demes) == 5
        assert len(graph.pulses) == 0
        assert len(graph.migrations) == 0
        assert len([deme for deme in graph.demes if math.isinf(deme.start_time)])
        assert [t * 4 * N0 for t in [T1, T2, T3, T4, math.inf]] == sorted(
            [deme.start_time for deme in graph.demes]
        )
        assert [0, 0, 0, 0, 0] == [deme.end_time for deme in graph.demes]
        assert [1, 1, 1, 1, 1] == [len(deme.epochs) for deme in graph.demes]
        assert graph["deme1"].ancestors == ["deme2"]
        assert graph["deme2"].ancestors == ["deme3"]
        assert graph["deme3"].ancestors == ["deme4"]
        assert graph["deme4"].ancestors == ["deme5"]
        assert graph["deme5"].ancestors == []

        for bad_t in (-1, 0, math.inf):
            cmd = f"-I 2 1 1 -ej {bad_t} 1 2"
            with pytest.raises(ValueError):
                demes.from_ms(cmd, N0=1)

        for bad_cmd in (
            # deme doesn't exist
            "-ej 1.0 1 2",
            "-ej 1.0 2 1",
            "-ej 1.0 2 3",
            "-I 3 0 0 2 -ej 1.0 1 4",
            "-I 3 0 0 2 -ej 1.0 4 1",
            "-I 3 0 0 2 -ej 1.0 4 10",
            # can't split a deme from itself
            "-ej 1.0 1 1",
            "-I 3 0 0 2 -ej 1.0 3 3",
            # can't split a deme after it's already split
            "-I 3 0 0 2 -ej 1.0 1 2 -ej 2.0 1 2",
            "-I 3 0 0 2 -ej 1.0 1 2 -ej 2.0 1 3",
        ):
            with pytest.raises(ValueError):
                demes.from_ms(bad_cmd, N0=1)

    def test_admixture_and_split(self):
        # Test that -ej and -es work together.
        N0 = 100
        T1, T2 = 1.0, 2.0
        p = 0.1
        cmd = f"-es {T1} 1 {p} -ej {T2} 2 1"
        graph = demes.from_ms(cmd, N0=N0)
        assert len(graph.demes) == 2
        assert len(graph.pulses) == 1
        assert len(graph.migrations) == 0
        # check deme1
        assert graph["deme1"].ancestors == []
        assert math.isinf(graph["deme1"].start_time)
        assert graph["deme1"].end_time == 0
        assert len(graph["deme1"].epochs) == 1
        # check deme2
        assert graph["deme2"].ancestors == ["deme1"]
        assert math.isclose(graph["deme2"].start_time, T2 * 4 * N0)
        assert math.isclose(graph["deme2"].end_time, T1 * 4 * N0)
        assert len(graph["deme2"].epochs) == 1
        # check pulse
        pulse = graph.pulses[0]
        assert pulse.source == "deme2"
        assert pulse.dest == "deme1"
        assert pulse.proportion == 1 - p
        assert math.isclose(pulse.time, T1 * 4 * N0)

    def test_args_from_file(self):
        # -f filename

        def writeparams(filename, params):
            with open(filename, "w") as f:
                print(params, file=f)

        def check_im(graph, *, num_demes, M, N0):
            assert len(graph.demes) == num_demes
            assert len(graph.migrations) == num_demes * (num_demes - 1)
            rate = M / (4 * N0)
            for migration in graph.migrations:
                assert math.isclose(migration.rate, rate / (num_demes - 1))

        def check_epochs(graph, *, alpha, T, N0):
            for deme in graph.demes:
                assert len(deme.epochs) == 2
                e1, e2 = deme.epochs
                size = N0 * math.exp(-alpha * T)
                assert math.isinf(e1.start_time)
                assert math.isclose(e1.end_time, T * 4 * N0)
                assert math.isclose(e2.start_time, T * 4 * N0)
                assert math.isclose(e2.end_time, 0)
                assert math.isclose(e1.start_size, size)
                assert math.isclose(e1.end_size, size)
                assert math.isclose(e2.start_size, size)
                assert math.isclose(e2.end_size, N0)

        N0 = 100
        T = 1.0
        M = 0.1
        alpha = 0.01
        with tempfile.TemporaryDirectory() as tmpdir:
            paramfile = str(pathlib.Path(tmpdir) / "params.txt")
            writeparams(paramfile, f"-I 2 1 1 {M}")
            # param file on its own
            cmd = f"-f {paramfile}"
            graph = demes.from_ms(cmd, N0=N0)
            check_im(graph, num_demes=2, M=M, N0=N0)
            # other params before -f
            cmd = f"-eG 0 {alpha} -eG {T} 0 -f {paramfile}"
            graph = demes.from_ms(cmd, N0=N0)
            check_im(graph, num_demes=2, M=M, N0=N0)
            check_epochs(graph, alpha=alpha, T=T, N0=N0)
            # params after -f
            cmd = f"-f {paramfile} -eG 0 {alpha} -eG {T} 0 "
            graph = demes.from_ms(cmd, N0=N0)
            check_im(graph, num_demes=2, M=M, N0=N0)
            check_epochs(graph, alpha=alpha, T=T, N0=N0)
            # params before and after -f
            cmd = f"-eG 0 {alpha} -f {paramfile} -eG {T} 0 "
            graph = demes.from_ms(cmd, N0=N0)
            check_im(graph, num_demes=2, M=M, N0=N0)
            check_epochs(graph, alpha=alpha, T=T, N0=N0)
            # -f inside a file specified with -f
            paramfile2 = str(pathlib.Path(tmpdir) / "params2.txt")
            writeparams(paramfile2, f"-eG 0 {alpha} -eG {T} 0 -f {paramfile}")
            cmd = f"-f {paramfile2}"
            graph = demes.from_ms(cmd, N0=N0)
            check_im(graph, num_demes=2, M=M, N0=N0)
            check_epochs(graph, alpha=alpha, T=T, N0=N0)

        # nonexistent file
        cmd = "-f nonexistent"
        with pytest.raises(OSError):
            demes.from_ms(cmd, N0=1)

    def test_bad_arguments(self):
        # Check that giving the wrong number of options is an error.
        for bad_cmd in (
            "-f",
            "-G",
            "-I",
            "-I 2 1 1 -g",
            "-I 2 1 1 -m",
            "-I 2 1 1 -ma",
            "-I 2 1 1 -eG",
            "-I 2 1 1 -eg",
            "-I 2 1 1 -eN",
            "-I 2 1 1 -en",
            "-I 2 1 1 -eM",
            "-I 2 1 1 -em",
            "-I 2 1 1 -ema",
            "-I 2 1 1 -es",
            "-I 2 1 1 -ej",
        ):
            with pytest.raises(ValueError):
                demes.from_ms(bad_cmd, N0=1)

    ##
    # Examples from the ms manual (November 10 2018).

    def test_msdoc_example1(self):
        # Instantaneous population size changes.
        cmd = "ms 15 1000 -t 2.0 -eN 1.0 .1 -eN 2.0 4.0"
        graph = demes.from_ms(cmd, N0=1)
        assert len(graph.migrations) == 0
        assert len(graph.pulses) == 0
        assert len(graph.demes) == 1
        assert len(graph.demes[0].epochs) == 3
        e1, e2, e3 = graph.demes[0].epochs
        assert e1.start_size == e1.end_size
        assert e2.start_size == e2.end_size
        assert e3.start_size == e3.end_size
        assert math.isclose(e2.end_size, 0.1 * e3.end_size)
        assert math.isclose(e1.end_size, 4 * e3.end_size)

    def test_msdoc_example2(self):
        # Generating an outgroup sequence.
        cmd = "ms 11 3 -t 2.0 -I 2 1 10 -ej 6.0 1 2"
        N0 = self.N0(theta=2, mu=1, length=1)
        graph = demes.from_ms(cmd, N0=N0)
        assert len(graph.migrations) == 0
        assert len(graph.pulses) == 0
        assert len(graph.demes) == 2
        d1, d2 = graph["deme1"], graph["deme2"]
        assert len(d1.epochs) == 1
        assert len(d2.epochs) == 1
        assert len(d1.ancestors) == 1
        assert len(d2.ancestors) == 0
        assert d2.name in d1.ancestors
        assert math.isclose(24 * N0, d1.start_time)

    def test_msdoc_example3(self):
        # Instantaneous size change followed by exponential growth.
        # Figure 1.
        cmd = "ms 15 1000 -t 6.4 -G 6.93 -eG 0.2 0.0 -eN 0.3 0.5"
        graph = demes.from_ms(cmd, N0=self.N0(theta=6.4, mu=1e-8, length=8000))
        assert len(graph.migrations) == 0
        assert len(graph.pulses) == 0
        assert len(graph.demes) == 1
        assert len(graph.demes[0].epochs) == 3

        N1, N2, N3 = 10000, 5000, 20000
        T1, T2 = 16000, 24000
        e1, e2, e3 = graph.demes[0].epochs
        assert math.isclose(e1.start_size, N1)
        assert math.isclose(e1.end_size, N1)
        # Note: coarse rounding used for the growth rate in the manual.
        assert math.isclose(e2.start_size, N2, rel_tol=1e-3)
        assert math.isclose(e2.end_size, N2, rel_tol=1e-2)
        assert math.isclose(e3.start_size, N2, rel_tol=1e-3)
        assert math.isclose(e3.end_size, N3)
        assert math.isclose(e1.end_time, T2)
        assert math.isclose(e2.end_time, T1)
        assert math.isclose(e3.end_time, 0)

    def test_msdoc_example4(self):
        # Two species with population size differences.
        # Figure 2.

        def verify(graph):
            N1, N2, N3, N4 = 10000, 2000, 5000, 40000
            T1, T2, T3 = 5000, 10000, 15000
            rel_tol = 1e-4

            assert len(graph.migrations) == 0
            assert len(graph.pulses) == 0
            assert len(graph.demes) == 2
            d1, d2 = graph["deme1"], graph["deme2"]
            assert math.isinf(d1.start_time)
            assert d1.end_time == 0
            assert math.isclose(d2.start_time, T3)
            assert d2.end_time == 0

            assert len(d1.epochs) == 2
            d1e1, d1e2 = d1.epochs
            assert math.isclose(d1e1.start_size, N1, rel_tol=rel_tol)
            assert math.isclose(d1e1.end_size, N1, rel_tol=rel_tol)
            assert math.isclose(d1e2.start_size, N1, rel_tol=rel_tol)
            assert math.isclose(d1e2.end_size, N4, rel_tol=rel_tol)
            assert math.isinf(d1e1.start_time)
            assert math.isclose(d1e1.end_time, T1)
            assert math.isclose(d1e2.start_time, T1)
            assert math.isclose(d1e2.end_time, 0)

            assert len(d2.epochs) == 2
            d2e1, d2e2 = d2.epochs
            assert math.isclose(d2e1.start_size, N2, rel_tol=rel_tol)
            assert math.isclose(d2e1.end_size, N2, rel_tol=rel_tol)
            assert math.isclose(d2e2.start_size, N3, rel_tol=rel_tol)
            assert math.isclose(d2e2.end_size, N3, rel_tol=rel_tol)
            assert math.isclose(d2e1.start_time, T3)
            assert math.isclose(d2e1.end_time, T2)
            assert math.isclose(d2e2.start_time, T2)
            assert math.isclose(d2e2.end_time, 0)

        mu = 1e-8
        sequence_length = 7000
        cmd1 = (
            "ms 15 100 "
            "-t 11.2 "
            "-I 2 3 12 "
            "-g 1 44.36 "
            "-n 2 0.125 "
            "-eg 0.03125 1 0.0 "
            "-en 0.0625 2 0.05 "
            "-ej 0.09375 2 1"
        )
        graph1 = demes.from_ms(
            cmd1, N0=self.N0(theta=11.2, mu=mu, length=sequence_length)
        )
        verify(graph1)

        # A second command is provided in the manual, using a different N0,
        # which should result in an identical model.
        cmd2 = (
            "ms 15 100 "
            "-t 2.8 "
            "-I 2 3 12 "
            "-g 1 11.09 "
            "-n 1 4.0 "
            "-n 2 0.5 "
            "-eg 0.125 1 0.0 "
            "-en 0.25 2 .2 "
            "-ej 0.375 2 1"
        )
        graph2 = demes.from_ms(
            cmd2, N0=self.N0(theta=2.8, mu=mu, length=sequence_length)
        )
        verify(graph2)

        graph1.assert_close(graph2)

    def test_msdoc_example5(self):
        # Stepping stone model with recent barrier.
        # Figure 3.
        cmd = (
            "ms 15 100 "
            "-t 3.0 "
            "-I 6 0 7 0 0 8 0 "
            "-m 1 2 2.5 "
            "-m 2 1 2.5 "
            "-m 2 3 2.5 "
            "-m 3 2 2.5 "
            "-m 4 5 2.5 "
            "-m 5 4 2.5 "
            "-m 5 6 2.5 "
            "-m 6 5 2.5 "
            "-em 2.0 3 4 2.5 "
            "-em 2.0 4 3 2.5"
        )
        N0 = self.N0(theta=3, mu=1e-8, length=1)
        graph = demes.from_ms(cmd, N0=N0)
        assert len(graph.pulses) == 0
        assert len(graph.migrations) == 10
        assert len(graph.demes) == 6

        def get_migrations(graph, source, dest):
            """Return a list of the migrations from source to dest."""
            migrations = []
            for migration in graph.migrations:
                if migration.source == source and migration.dest == dest:
                    migrations.append(migration)
            return migrations

        migration_rate = 2.5 / (4 * N0)
        # time of migration barrier
        T1 = 2.0 * 4 * N0

        for d1, d2 in [
            ("deme1", "deme2"),
            ("deme2", "deme3"),
            ("deme4", "deme5"),
            ("deme5", "deme6"),
        ]:
            m12 = get_migrations(graph, d1, d2)
            m21 = get_migrations(graph, d2, d1)
            assert len(m12) == len(m21) == 1
            m12 = m12[0]
            m21 = m21[0]
            assert math.isclose(m12.rate, migration_rate)
            assert math.isclose(m21.rate, migration_rate)
            assert math.isinf(m12.start_time)
            assert math.isinf(m21.start_time)
            assert math.isclose(m12.end_time, 0)
            assert math.isclose(m21.end_time, 0)

        m34 = get_migrations(graph, "deme3", "deme4")
        m43 = get_migrations(graph, "deme4", "deme3")
        assert len(m34) == len(m43) == 1
        m34 = m34[0]
        m43 = m43[0]
        assert math.isclose(m34.rate, migration_rate)
        assert math.isclose(m43.rate, migration_rate)
        assert math.isinf(m34.start_time)
        assert math.isinf(m43.start_time)
        assert math.isclose(m34.end_time, T1)
        assert math.isclose(m43.end_time, T1)

    ##
    # Ms commands from publications.

    def test_zigzag(self):
        # Schiffels & Durbin (2014), https://doi.org/10.1038/ng.3015
        # Ms command from section 7 of the supplement (pg. 25).
        cmd = (
            "ms 4 1 "
            "-t 7156.0000000 "
            "-r 2000.0000 10000000 "
            "-eN 0 5 "
            "-eG 0.000582262 1318.18 "
            "-eG 0.00232905 -329.546 "
            "-eG 0.00931619 82.3865 "
            "-eG 0.0372648 -20.5966 "
            "-eG 0.149059 5.14916 "
            "-eN 0.596236 0.5 "
            "-T"
        )
        mu = 1.25e-8
        theta = 7156
        sequence_length = 10000000
        graph = demes.from_ms(
            cmd, N0=self.N0(theta=theta, mu=mu, length=sequence_length)
        )
        assert len(graph.migrations) == 0
        assert len(graph.pulses) == 0
        assert len(graph.demes) == 1
        assert len(graph.demes[0].epochs) == 7

        # Compare with zigzag in examples folder.
        zz = demes.load(example_dir / "zigzag.yml")
        zz.generation_time = None
        zz.demes[0].name = "deme1"
        # We use a lower tolerance than the default because the example model
        # uses somewhat arbitrary rounding.
        graph.assert_close(zz, rel_tol=1e-4)

    def test_gutenkunst_ooa(self):
        # Gutenkunst et al. (2009), https://doi.org/10.1371/journal.pgen.1000695
        # Ms command from section 5.2 of the supplement (pg. 10--12).
        cmd = (
            "-I 3 1 1 1 "
            "-n 1 1.682020 "
            "-n 2 3.736830 "
            "-n 3 7.292050 "
            "-eg 0 2 116.010723 "
            "-eg 0 3 160.246047 "
            "-ma x 0.881098 0.561966 0.881098 x 2.797460 0.561966 2.797460 x "
            "-ej 0.028985 3 2 "
            "-en 0.028985 2 0.287184 "
            "-ema 0.028985 3 x 7.293140 x 7.293140 x x x x x "
            "-ej 0.197963 2 1 "
            "-en 0.303501 1 1"
        )
        graph = demes.from_ms(
            cmd,
            N0=7310,
            deme_names=["YRI", "CEU", "CHB"],
        )
        assert len(graph.demes) == 3
        assert len(graph.migrations) == 8
        # TODO: more checks.
        # We don't compare to the gutenkunst_ooa.yml file in the examples
        # folder because that graph introduces additional demes for the
        # ancestral and OOA populations which are not present in the original
        # ms-based demographic model.

    ##
    # api stuff

    def test_deme_names(self):
        graph = demes.from_ms("-I 2 1 1 1.0", N0=1, deme_names=["A", "B"])
        assert len(graph.demes) == 2
        assert "A" in graph
        assert "B" in graph
        for migration in graph.migrations:
            assert migration.source in ["A", "B"]
            assert migration.dest in ["A", "B"]

        graph = demes.from_ms("-es 1.0 1 0.1", N0=1, deme_names=["A", "B"])
        assert len(graph.demes) == 2
        assert "A" in graph
        assert "B" in graph
        assert len(graph.pulses) == 1
        assert graph.pulses[0].source == "B"
        assert graph.pulses[0].dest == "A"

        # bad deme names
        for bad_deme_names in (
            ["A"],
            ["A", "A"],
            ["A", "B", "C"],
        ):
            with pytest.raises(ValueError):
                demes.from_ms("-I 2 1 1", N0=1, deme_names=bad_deme_names)

    def test_unhandled_event(self):
        from demes.ms import parse_ms_args, build_graph

        class Foo:
            t: float

        foo = Foo()
        foo.t = 0
        args, _ = parse_ms_args("-I 2 1 1")
        args.initial_state.append(foo)
        with pytest.raises(AssertionError):
            build_graph(args, 1)

        foo.t = 1.0
        args, _ = parse_ms_args("-I 2 1 1")
        args.initial_state.append(foo)



import math
import itertools
import collections

import numpy as np
import hypothesis as hyp
import hypothesis.strategies as st

import demes


@st.composite
def deme_names(draw, max_length=20):
    """
    A hypothesis strategy for creating a valid deme name.
    """
    name = draw(st.text(min_size=1, max_size=max_length))
    # Names must be valid Python identifiers.
    hyp.assume(name.isidentifier())
    return name


@st.composite
def yaml_strings(draw, min_size=1, max_size=100):
    """
    A hypothesis strategy for creating a valid YAML string.

    From https://yaml.org/spec/1.2/spec.html#id2770814

        To ensure readability, YAML streams use only the printable subset of
        the Unicode character set. The allowed character range explicitly
        excludes the C0 control block #x0-#x1F (except for TAB #x9, LF #xA,
        and CR #xD which are allowed), DEL #x7F, the C1 control block #x80-#x9F
        (except for NEL #x85 which is allowed), the surrogate block #xD800-#xDFFF,
        #xFFFE, and #xFFFF.

        On input, a YAML processor must accept all Unicode characters except
        those explicitly excluded above.

        On output, a YAML processor must only produce acceptable characters.
        Any excluded characters must be presented using escape sequences.
        In addition, any allowed characters known to be non-printable should
        also be escaped. This isn’t mandatory since a full implementation would
        require extensive character property tables.
    """
    return draw(
        st.text(
            alphabet=st.characters(
                blacklist_categories=(
                    "Cc",  # control block (C0 and C1)
                    "Cs",  # surrogate block
                ),
                blacklist_characters=("\ufffe", "\uffff"),
                whitelist_characters=("\x09", "\x0a", "\x0d", "\x85"),
            ),
            min_size=min_size,
            max_size=max_size,
        )
    )


@st.composite
def epochs_lists(draw, start_time=math.inf, max_epochs=5):
    """
    A hypothesis strategy for creating lists of Epochs for a deme.

    :param float start_time: The start time of the deme.
    :param int max_epochs: The maximum number of epochs in the list.
    """
    assert max_epochs >= 2
    times = draw(
        st.lists(
            st.floats(min_value=0, max_value=start_time, exclude_max=True),
            unique=True,
            min_size=1,
            max_size=max_epochs,
        )
    )
    times.sort(reverse=True)
    epochs = []

    for i, end_time in enumerate(times):
        start_size = draw(
            st.floats(min_value=0, exclude_min=True, allow_infinity=False)
        )
        if i == 0 and math.isinf(start_time):
            end_size = start_size
        else:
            end_size = draw(
                st.floats(min_value=0, exclude_min=True, allow_infinity=False)
            )
        cloning_rate = draw(st.floats(min_value=0, max_value=1))
        selfing_rate = draw(st.floats(min_value=0, max_value=1 - cloning_rate))

        epochs.append(
            dict(
                end_time=end_time,
                start_size=start_size,
                end_size=end_size,
                cloning_rate=cloning_rate,
                selfing_rate=selfing_rate,
            )
        )

    return epochs


@st.composite
def migration_matrices(
    draw, graph, max_migrations=10, max_additional_migration_intervals=5
):
    """
    A hypothesis strategy for creating migration matrices for a graph.
    """
    n = len(graph.demes)
    assert n > 0

    uniq_deme_times = set(deme.start_time for deme in graph.demes)
    uniq_deme_times.update(deme.end_time for deme in graph.demes)
    start_time, *end_times = sorted(uniq_deme_times, reverse=True)

    # Identify the first time at which 2 or more demes exist simultaneously.
    for end_time in end_times:
        if sum(1 for deme in graph.demes if deme.start_time <= start_time) > 1:
            break
        start_time = end_time

    if start_time == end_times[-1]:
        # No two demes exist simultaneously.
        return [[[0] * n for _ in range(n)]], math.inf, [0]

    saved_start_time = start_time

    # Partition time intervals even further.
    additional_times = draw(
        st.lists(
            st.floats(min_value=end_times[-1], max_value=start_time, exclude_max=True),
            unique=True,
            min_size=0,
            max_size=max_additional_migration_intervals,
        )
    )
    end_times = sorted(set(end_times + additional_times), reverse=True)

    mm_list = [[[0] * n for _ in range(n)] for _ in range(len(end_times))]
    n_migrations = draw(st.integers(min_value=0, max_value=max_migrations))

    for migration_matrix, end_time in zip(mm_list, end_times):
        # Find demes alive in this interval.
        deme_indices = [
            j
            for j, deme in enumerate(graph.demes)
            if (
                deme.start_time >= start_time > deme.end_time
                and deme.start_time > end_time >= deme.end_time
            )
        ]
        if len(deme_indices) < 2:
            continue

        # Select pairs of demes for migration.
        pairs = list(itertools.permutations(deme_indices, 2))
        pair_indices = draw(
            st.lists(
                st.integers(min_value=0, max_value=len(pairs) - 1),
                unique=True,
                min_size=0,
                max_size=min(len(pairs), n_migrations),
            )
        )

        for k in pair_indices:
            a, b = pairs[k]
            assert migration_matrix[a][b] == 0
            max_rate = 1 - sum(migration_matrix[a])
            if math.isclose(max_rate, 0):
                continue
            n_migrations -= 1
            rate = draw(st.floats(min_value=0, max_value=max_rate, exclude_min=True))
            migration_matrix[a][b] = rate

        if n_migrations == 0:
            break
        start_time = end_time

    return mm_list, saved_start_time, end_times


@st.composite
def migrations_lists(draw, graph, max_migrations=10):
    """
    A hypothesis strategy for creating a migration list for a graph.
    """
    mm_list, start_time, end_times = draw(
        migration_matrices(graph, max_migrations=max_migrations)
    )
    assert len(mm_list) == len(end_times)
    migrations = []
    for migration_matrix, end_time in zip(mm_list, end_times):
        for j, row in enumerate(migration_matrix):
            for k, rate in enumerate(row):
                if rate > 0:
                    migration = demes.AsymmetricMigration(
                        source=graph.demes[k].name,
                        dest=graph.demes[j].name,
                        start_time=start_time,
                        end_time=end_time,
                        rate=rate,
                    )
                    migrations.append(migration)
        start_time = end_time
    return migrations


@st.composite
def pulses_lists(draw, graph, max_pulses=10):
    """
    A hypothesis strategy for creating a pulses list for a graph.
    """
    n_pulses = draw(st.integers(min_value=0, max_value=max_pulses))
    pulses = []
    ingress_proportions = collections.defaultdict(lambda: 0)
    for j, deme_j in enumerate(graph.demes[:-1]):
        for deme_k in graph.demes[j + 1 :]:
            time_lo = max(deme_j.end_time, deme_k.end_time)
            time_hi = min(deme_j.start_time, deme_k.start_time)

            # We wish to draw times for the pulses. They must be in the open
            # interval (time_lo, time_hi) to ensure the pulse doesn't happen
            # at any deme's start_time or end_time, which could be invalid.
            # So we check there is at least one floating point number between
            # time_lo and time_hi.
            if time_hi <= np.nextafter(time_lo, np.inf, dtype=float):
                continue
            n = draw(st.integers(min_value=0, max_value=n_pulses))
            for _ in range(n):
                source, dest = deme_j.name, deme_k.name
                if draw(st.booleans()):
                    source, dest = dest, source
                time = draw(
                    st.floats(
                        min_value=time_lo,
                        max_value=time_hi,
                        exclude_min=True,
                        exclude_max=True,
                    )
                )
                max_proportion = 1 - ingress_proportions[(dest, time)]
                if math.isclose(max_proportion, 0):
                    continue
                proportion = draw(
                    st.floats(
                        min_value=0,
                        max_value=max_proportion,
                        exclude_min=True,
                        exclude_max=True,
                    )
                )
                ingress_proportions[(dest, time)] += proportion
                pulse = dict(
                    source=source,
                    dest=dest,
                    time=time,
                    proportion=proportion,
                )
                pulses.append(pulse)
                n_pulses -= 1
            if n_pulses == 0:
                break
        if n_pulses == 0:
            break
    return pulses


@st.composite
def graphs(draw, max_demes=5, max_epochs=10, max_migrations=10, max_pulses=10):
    """
    A hypothesis strategy for creating a Graph.

    .. code-block::

        @hypothesis.given(graphs())
        def test_something(self, graph: demes.Graph):
            # Do something with the ``graph``.
            ...

    :param int max_demes: The maximum number of demes in the graph.
    :param int max_epochs: The maximum number of epochs per deme.
    :param int max_migrations: The maximum number of migrations in the graph.
    :param int max_pulses: The maximum number of pulses in the graph.
    """
    generation_time = draw(st.none() | st.floats(min_value=1e-9, max_value=1e6))
    if generation_time is None:
        time_units = "generations"
    else:
        time_units = draw(yaml_strings())
    b = demes.Builder(
        description=draw(yaml_strings()),
        generation_time=generation_time,
        time_units=time_units,
        doi=draw(st.lists(yaml_strings(), max_size=3)),
    )

    for deme_name in draw(st.sets(deme_names(), min_size=1, max_size=max_demes)):
        ancestors = []
        proportions = []
        start_time = math.inf
        n_demes = 0 if "demes" not in b.data else len(b.data["demes"])
        if n_demes > 0:
            # draw indices into demes list to use as ancestors
            anc_idx = draw(
                st.lists(
                    st.integers(min_value=0, max_value=n_demes - 1),
                    unique=True,
                    max_size=n_demes,
                )
            )
            if len(anc_idx) > 0:
                time_hi = min(b.data["demes"][j]["start_time"] for j in anc_idx)
                time_lo = max(
                    b.data["demes"][j]["epochs"][-1]["end_time"] for j in anc_idx
                )
                # If time_hi > time_lo, the proposed ancestors exist
                # at the same time. So we draw a number for the deme's
                # start_time, which must be in the half-open interval
                # [time_lo, time_hi), with the further constraint that the
                # start_time cannot be 0.
                # However, there may not be any floating point numbers between
                # 0 and time_hi even if time_hi > 0, so we check that time_hi
                # is greater than the smallest positive number.
                if (time_lo > 0 and time_hi > time_lo) or (
                    time_lo == 0 and time_hi > np.finfo(float).tiny
                ):
                    # Draw a start time and the ancestry proportions.
                    start_time = draw(
                        st.floats(
                            min_value=time_lo,
                            max_value=time_hi,
                            exclude_max=True,
                            # Can't have start_time=0.
                            exclude_min=time_lo == 0,
                        )
                    )
                    ancestors = [b.data["demes"][j]["name"] for j in anc_idx]
                    if len(ancestors) == 1:
                        proportions = [1.0]
                    else:
                        proportions = draw(
                            st.lists(
                                st.integers(min_value=1, max_value=10 ** 9),
                                min_size=len(ancestors),
                                max_size=len(ancestors),
                            )
                        )
                        psum = sum(proportions)
                        proportions = [p / psum for p in proportions]
        b.add_deme(
            name=deme_name,
            description=draw(st.none() | yaml_strings()),
            ancestors=ancestors,
            proportions=proportions,
            epochs=draw(epochs_lists(start_time=start_time, max_epochs=max_epochs)),
            start_time=start_time,
        )

    graph = b.resolve()
    graph.migrations = draw(migrations_lists(graph, max_migrations=max_migrations))
    graph.pulses = draw(pulses_lists(graph, max_pulses=max_pulses))
    # Resolve the graph again. This is not strictly necessary, but has only
    # a small computational overhead and serves to catch simple errors in
    # the migrations_lists()/pulses_lists() implementations.
    graph = demes.Builder.fromdict(graph.asdict()).resolve()
    return graph



import copy
import math
import argparse
import logging
import sys
from typing import Any, Dict, List, Mapping, MutableMapping, Set, Tuple

import attr

import demes
from .demes import finite, positive, non_negative, unit_interval

logger = logging.getLogger(__name__)


class ValueErrorArgumentParser(argparse.ArgumentParser):
    """
    An AgumentParser that throws a ValueError instead of exiting.
    It would be preferred to catch and reraise TypeError and ValueError
    separately, but that isn't possible without touching ArgumentParser
    internals because those errors are caught by parse_known_args(),
    which then throws an argparse.ArgumentError.
    So we just use ValueError for everything.
    """

    def error(self, message):
        _, exc, traceback = sys.exc_info()
        raise ValueError(str(exc))


def coerce_nargs(obj_creator, append=False):
    """
    Return an argparse action that coerces the collected nargs into
    data class using the specified class/function ``obj_creator``.
    This simplifies the validation of heterogeneously-typed args lists.
    If ``append`` is True, the option may be specified multiple times
    and each coerced object will be appended to a list.

    .. code::

        @attr.define
        class Foo
            n = attr.ib(converter=int)
            x = attr.ib(converter=float)

        parser.add_argument(
            "--foo",
            nargs=2,
            action=coerce_nargs(Foo),
            ...
        )

        args = parser.parse_args("--foo 6 1e-5")
        assert isinstance(args.foo, Foo)

    """
    parent_class = argparse.Action
    if append:
        parent_class = argparse._AppendAction

    class CoerceAction(parent_class):
        def __call__(self, parser, namespace, values, *args, **kwargs):
            obj = obj_creator(*values)
            # We save the option strings to later identify which option was
            # used to create the object. E.g. -n and -en options both map to
            # PopulationSizeChange, but have slightly different semantics.
            obj.option_strings = self.option_strings
            if append:
                super().__call__(parser, namespace, obj, *args, **kwargs)
            else:
                setattr(namespace, self.dest, obj)

    return CoerceAction


##
# Data classes for ms options. These help to separate input validation
# from the graph building procedure.


@attr.define
class Option:
    # This attribute is set by CoerceAction.
    option_strings: List[str] = attr.ib(init=False)


@attr.define
class Structure(Option):
    # -I npop n1 n2 ... [4*N0*m]
    npop = attr.ib(converter=int, validator=positive)
    n: Any = attr.ib()  # samples list: currently ignored
    rate = attr.ib(converter=float, validator=non_negative)

    def __attrs_post_init__(self):
        if len(self.n) != self.npop:
            raise ValueError("sample configuration doesn't match number of demes")

    @classmethod
    def from_nargs(cls, *args):
        npop, *n = args
        rate = 0
        try:
            npop = int(npop)
        except ValueError:
            raise ValueError(f"-I 'npop' ({args[0]}) not an integer")
        if len(n) == npop + 1:
            *n, rate = n
        return cls(npop, n, rate)


@attr.define
class Event(Option):
    # The first param of every demographic event is the event time.
    # For ms options without a time parameter (i.e. those that set the
    # initial simulation state), t is just set to zero.
    t = attr.ib(converter=float, validator=non_negative)


@attr.define
class GrowthRateChange(Event):
    # -G α
    # -eG t α
    alpha = attr.ib(converter=float, validator=finite)


@attr.define
class PopulationGrowthRateChange(Event):
    # -g i α
    # -eg t i α
    i = attr.ib(converter=int, validator=positive)
    alpha = attr.ib(converter=float, validator=finite)


@attr.define
class SizeChange(Event):
    # -eN t x
    x = attr.ib(converter=float, validator=non_negative)


@attr.define
class PopulationSizeChange(Event):
    # -n i x
    # -en t i x
    i = attr.ib(converter=int, validator=positive)
    x = attr.ib(converter=float, validator=non_negative)


@attr.define
class MigrationRateChange(Event):
    # -eM t x
    x = attr.ib(converter=float, validator=non_negative)


@attr.define
class MigrationMatrixEntryChange(Event):
    # -m i j rate
    # -em t i j rate
    i = attr.ib(converter=int, validator=positive)
    j = attr.ib(converter=int, validator=positive)
    rate = attr.ib(converter=float, validator=non_negative)


@attr.define
class MigrationMatrixChange(Event):
    # -ma M11 M12 M12 ... M21 ...
    # -ema t npop M11 M12 M12 ... M21 ...
    npop = attr.ib(converter=int, validator=positive)
    mm_vector = attr.ib()

    @property
    def M(self):
        """
        Convert the args vector into a square list-of-lists matrix.
        """
        if len(self.mm_vector) != self.npop ** 2:
            raise ValueError(
                f"Must be npop^2={self.npop**2} migration matrix entries: "
                f"{self.mm_vector}"
            )
        migration_matrix = [[0 for j in range(self.npop)] for k in range(self.npop)]
        for j in range(self.npop):
            for k in range(self.npop):
                if j != k:
                    rate = self.mm_vector[j * self.npop + k]

                    # Convert to float if possible. Ms ignores migration matrix
                    # diagonals, as well as migration entries for "joined" demes.
                    # The manual suggests to indicate diagonal elements with:
                    #       x's, or any symbol one chooses to make the matrix
                    #       more readable.
                    # NaNs will be caught later during graph resolution if we
                    # really attempt to use the value.
                    try:
                        rate = float(rate)
                    except ValueError:
                        rate = math.nan

                    migration_matrix[j][k] = rate
        return migration_matrix

    @classmethod
    def from_nargs(cls, *args):
        t, npop, *mm_vector = args
        return cls(t, npop, mm_vector)


@attr.define
class Admixture(Event):
    # -es t i p
    i = attr.ib(converter=int, validator=positive)
    p = attr.ib(converter=float, validator=unit_interval)


@attr.define
class PopulationSplit(Event):
    # -ej t i j
    i = attr.ib(converter=int, validator=positive)
    j = attr.ib(converter=int, validator=positive)


def parse_ms_args(command: str):
    parser = ValueErrorArgumentParser()

    class LoadFromFile(argparse.Action):
        def __call__(self, parser, namespace, filename, option_string=None):
            # parse arguments in the file and store them in the target namespace
            with open(filename) as f:
                args = f.read().split()
            parser.parse_args(args, namespace)

    parser.add_argument(
        "-f",
        "--filename",
        action=LoadFromFile,
        help="Insert commands from a file at this point in the command line.",
    )
    parser.add_argument(
        "--structure",
        "-I",
        nargs="+",
        action=coerce_nargs(Structure.from_nargs),
        metavar="value",
        help=(
            "Sample from populations with the specified deme structure. "
            "The arguments are of the form 'num_demes "
            "n1 n2 ... [4N0m]', specifying the number of populations, "
            "the sample configuration, and optionally, the migration "
            "rate for a symmetric island model"
        ),
    )
    parser.add_argument(
        "--population-size",
        "-n",
        nargs=2,
        action=coerce_nargs(lambda *x: PopulationSizeChange(0, *x), append=True),
        dest="initial_state",
        default=[],
        metavar=("i", "x"),
        help="Set the size of a specific population to size*N0.",
    )
    parser.add_argument(
        "--population-growth-rate",
        "-g",
        nargs=2,
        action=coerce_nargs(lambda *x: PopulationGrowthRateChange(0, *x), append=True),
        dest="initial_state",
        default=[],
        metavar=("i", "α"),
        help="Set the growth rate to alpha for a specific population.",
    )
    parser.add_argument(
        "--growth-rate",
        "-G",
        nargs=1,
        action=coerce_nargs(lambda *x: GrowthRateChange(0, *x), append=True),
        dest="initial_state",
        default=[],
        metavar="α",
        help="Set the growth rate to alpha for all populations.",
    )
    parser.add_argument(
        "--migration-matrix-entry",
        "-m",
        nargs=3,
        action=coerce_nargs(lambda *x: MigrationMatrixEntryChange(0, *x), append=True),
        dest="initial_state",
        default=[],
        metavar=("i", "j", "rate"),
        help=(
            "Sets an entry M[i, j] in the migration matrix to the "
            "specified rate. i and j are (1-indexed) population "
            "IDs. Multiple options can be specified."
        ),
    )
    parser.add_argument(
        "--migration-matrix",
        "-ma",
        nargs="+",
        action=coerce_nargs(
            lambda *x: MigrationMatrixChange.from_nargs(0, 1, *x), append=True
        ),
        dest="initial_state",
        default=[],
        metavar="entry",
        help=(
            "Sets the migration matrix to the specified value. The "
            "entries are in the order M[1, 1], M[1, 2], ..., M[2, 1], "
            "M[2, 2], ..., M[N, N], where N is the number of populations. "
            "Diagonal entries may be written as 'x'."
        ),
    )
    parser.add_argument(
        "--growth-rate-change",
        "-eG",
        nargs=2,
        action=coerce_nargs(GrowthRateChange, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "alpha"),
        help="Set the growth rate for all populations to alpha at time t",
    )
    parser.add_argument(
        "--population-growth-rate-change",
        "-eg",
        nargs=3,
        action=coerce_nargs(PopulationGrowthRateChange, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "i", "alpha"),
        help=("Set the growth rate for a specific population to " "alpha at time t"),
    )
    parser.add_argument(
        "--size-change",
        "-eN",
        nargs=2,
        action=coerce_nargs(SizeChange, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "x"),
        help="Set the population size for all populations to x * N0 at time t",
    )
    parser.add_argument(
        "--population-size-change",
        "-en",
        nargs=3,
        action=coerce_nargs(PopulationSizeChange, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "i", "x"),
        help=(
            "Set the population size for a specific population to " "x * N0 at time t"
        ),
    )
    parser.add_argument(
        "--migration-rate-change",
        "-eM",
        nargs=2,
        action=coerce_nargs(MigrationRateChange, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "x"),
        help=(
            "Set the symmetric island model migration rate to "
            "x / (npop - 1) at time t"
        ),
    )
    parser.add_argument(
        "--migration-matrix-entry-change",
        "-em",
        action=coerce_nargs(MigrationMatrixEntryChange, append=True),
        dest="demographic_events",
        metavar=("time", "i", "j", "rate"),
        nargs=4,
        default=[],
        help=(
            "Sets an entry M[i, j] in the migration matrix to the "
            "specified rate at the specified time. i and j are "
            "(1-indexed) population IDs."
        ),
    )
    parser.add_argument(
        "--migration-matrix-change",
        "-ema",
        nargs="+",
        default=[],
        action=coerce_nargs(MigrationMatrixChange.from_nargs, append=True),
        dest="demographic_events",
        metavar="entry",
        help=(
            "Sets the migration matrix to the specified value at time t. "
            "The entries are in the order M[1, 1], M[1, 2], ..., M[2, 1], "
            "M[2, 2], ..., M[N, N], where N is the number of populations. "
            "Diagonal entries may be written as 'x'."
        ),
    )
    parser.add_argument(
        "--admixture",
        "-es",
        nargs=3,
        action=coerce_nargs(Admixture, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "i", "p"),
        help=(
            "Split the specified population into a new population, such "
            "that the specified proportion of lineages remains in "
            "the population i. Forwards in time this "
            "corresponds to an admixture event. The new population has ID "
            "num_demes + 1. Migration rates to and from the new "
            "population are set to 0, and growth rate is 0 and the "
            "population size for the new population is N0."
        ),
    )
    parser.add_argument(
        "--population-split",
        "-ej",
        nargs=3,
        action=coerce_nargs(PopulationSplit, append=True),
        dest="demographic_events",
        default=[],
        metavar=("t", "i", "j"),
        help=(
            "Move all lineages in population i to j at time t. "
            "Forwards in time, this corresponds to a population split "
            "in which lineages in j split into i. All migration "
            "rates for population i are set to zero."
        ),
    )

    args, unknown = parser.parse_known_args(command.split())
    # Sort demographic events args by the time field.
    args.demographic_events.sort(key=lambda x: x.t)
    return args, unknown


def demes_sorted_by_ancestry(demes_data: List[Mapping]):
    """
    Return the demes list sorted so that ancestors come before descendants.
    """
    # Iterate through the remaining demes and insert them once all
    # ancestors are already inserted.
    remaining = demes_data.copy()
    inserted = dict()
    n_remaining = len(remaining)
    while n_remaining > 0:
        for deme in remaining.copy():
            ancestors_in_graph = True
            for ancestor in deme.get("ancestors", []):
                if ancestor not in inserted:
                    ancestors_in_graph = False
                    break
            if ancestors_in_graph:
                remaining.remove(deme)
                inserted[deme["name"]] = deme
        if n_remaining == len(remaining):
            raise ValueError(
                "Cannot resolve graph because an ancestor/descendant cycle "
                f"exists among: {deme['name'] for deme in remaining}"
            )
        n_remaining = len(remaining)
    return list(inserted.values())


def migrations_from_mm_list(
    mm_list: List[List[List[float]]], end_times: List[float], deme_names: List[str]
) -> List[MutableMapping]:
    """
    Convert a list of migration matrices into a list of migration dicts.
    """
    assert len(mm_list) == len(end_times)
    migrations: List[MutableMapping] = []
    current: Dict[Tuple[int, int], MutableMapping] = dict()
    start_time = math.inf
    for migration_matrix, end_time in zip(mm_list, end_times):
        n = len(migration_matrix)
        assert n == len(deme_names)
        for j in range(n):
            assert n == len(migration_matrix[j])
            for k in range(n):
                if j == k:
                    continue
                rate = migration_matrix[j][k]
                mm = current.get((j, k))
                if mm is None:
                    if rate != 0:
                        mm = dict(
                            source=deme_names[j],
                            dest=deme_names[k],
                            start_time=start_time,
                            end_time=end_time,
                            rate=rate,
                        )
                        current[(j, k)] = mm
                        migrations.append(mm)
                else:
                    if rate == 0:
                        del current[(j, k)]
                    elif mm["rate"] == rate:
                        # extend mm
                        mm["end_time"] = end_time
                    else:
                        mm = dict(
                            source=deme_names[j],
                            dest=deme_names[k],
                            start_time=start_time,
                            end_time=end_time,
                            rate=rate,
                        )
                        current[(j, k)] = mm
                        migrations.append(mm)
        start_time = end_time
    return migrations


def build_graph(args, N0: float) -> demes.Graph:
    num_demes = 1
    # List of migration matrices in time-descending order (oldest to most recent).
    mm_list = [[[0.0]]]
    # End times for each migration matrix.
    mm_end_times = [0.0]
    # Indexes of demes that have been joined (-ej option).
    joined: Set[int] = set()

    if args.structure is not None:
        # -I npop n1 n2 .. [rate]
        num_demes = args.structure.npop
        if num_demes > 1:
            mm_list[0] = [
                [
                    args.structure.rate / (num_demes - 1) * int(j != k)
                    for j in range(num_demes)
                ]
                for k in range(num_demes)
            ]

    # Start building the demography.
    b = demes.Builder()
    for j in range(num_demes):
        initial_epoch = dict(end_size=N0, end_time=0)
        b.add_deme(f"deme{j + 1}", start_time=math.inf, epochs=[initial_epoch])

    def convert_population_id(population_id):
        """
        Checks the specified population ID makes sense and returns
        it as a zero-based index into the demes list.
        """
        if population_id < 1 or population_id > num_demes:
            raise ValueError(
                f"Bad population ID '{population_id}': "
                f"must be between 1 and num_demes ({num_demes})"
            )
        pid = population_id - 1
        if pid in joined:
            raise ValueError(
                f"Bad population ID '{population_id}': "
                "population previously joined with -ej"
            )
        return pid

    def epoch_resolve(deme, time):
        """
        Return the oldest epoch if it has end_time == time. If not, create a
        new oldest epoch with end_time=time. Also resolve sizes by dealing
        with the growth_rate attribute (if required).
        """
        epoch = deme["epochs"][0]
        start_time = deme["start_time"]
        end_time = epoch["end_time"]
        if not (start_time > time >= end_time):
            raise ValueError(
                f"time {time} outside {deme['name']}'s existence interval "
                f"(start_time={start_time}, end_time={end_time}]"
            )

        if time > end_time:
            new_epoch = copy.deepcopy(epoch)
            # find size at given time
            growth_rate = epoch.pop("growth_rate", 0)
            dt = time - epoch["end_time"]
            size_at_t = epoch["end_size"] * math.exp(-growth_rate * dt)
            epoch["start_size"] = size_at_t
            new_epoch["end_size"] = size_at_t
            new_epoch["end_time"] = time
            deme["epochs"].insert(0, new_epoch)
            epoch = new_epoch

        return epoch

    def migration_matrix_at(time):
        """
        Return the oldest migration matrix if it has end_time == time. If not,
        create a new oldest migration matrix with end_time = time.
        """
        migration_matrix = mm_list[0]
        if time > mm_end_times[0]:
            # We need a new migration matrix.
            migration_matrix = copy.deepcopy(migration_matrix)
            mm_list.insert(0, migration_matrix)
            mm_end_times.insert(0, time)
        return migration_matrix

    # Process the initial_state options followed by the demographic_events.
    for event in args.initial_state + args.demographic_events:
        time = 4 * N0 * event.t
        if isinstance(event, GrowthRateChange):
            # -G α
            # -eG t α
            growth_rate = event.alpha / (4 * N0)
            for deme in b.data["demes"]:
                current_epoch = deme["epochs"][0]
                current_growth_rate = current_epoch.get("growth_rate", 0)
                if current_growth_rate != growth_rate:
                    epoch = epoch_resolve(deme, time)
                    epoch["growth_rate"] = growth_rate

        elif isinstance(event, PopulationGrowthRateChange):
            # -g i α
            # -eg t i α
            pid = convert_population_id(event.i)
            growth_rate = event.alpha / (4 * N0)
            deme = b.data["demes"][pid]
            current_epoch = deme["epochs"][0]
            current_growth_rate = current_epoch.get("growth_rate", 0)
            if current_growth_rate != growth_rate:
                epoch = epoch_resolve(deme, time)
                epoch["growth_rate"] = growth_rate

        elif isinstance(event, SizeChange):
            # -eN t x
            size = event.x * N0
            for deme in b.data["demes"]:
                current_epoch = deme["epochs"][0]
                current_growth_rate = current_epoch.get("growth_rate", 0)
                if current_growth_rate != 0 or current_epoch["end_size"] != size:
                    epoch = epoch_resolve(deme, time)
                    epoch["growth_rate"] = 0
                    epoch["end_size"] = size

        elif isinstance(event, PopulationSizeChange):
            # -n i x
            # -en t i x
            pid = convert_population_id(event.i)
            size = event.x * N0
            deme = b.data["demes"][pid]
            current_epoch = deme["epochs"][0]
            current_growth_rate = current_epoch.get("growth_rate", 0)
            if current_growth_rate != 0 or current_epoch["end_size"] != size:
                epoch = epoch_resolve(deme, time)
                epoch["end_size"] = size
                # set growth_rate to 0 for -en option, but not for -n option
                if "-en" in event.option_strings:
                    epoch["growth_rate"] = 0

        elif isinstance(event, PopulationSplit):
            # -ej t i j
            pop_i = convert_population_id(event.i)
            pop_j = convert_population_id(event.j)

            b.data["demes"][pop_i]["start_time"] = time
            b.data["demes"][pop_i]["ancestors"] = [f"deme{pop_j + 1}"]

            mm = migration_matrix_at(time)
            # Turn off migrations to/from deme i.
            for k in range(num_demes):
                if k != pop_i:
                    mm[k][pop_i] = 0
                    mm[pop_i][k] = 0

            # Record pop_i so that this index isn't used by later events.
            joined.add(pop_i)

        elif isinstance(event, Admixture):
            # -es t i p
            pid = convert_population_id(event.i)

            # Add a new deme which will be the source of a migration pulse.
            new_pid = num_demes
            b.add_deme(
                f"deme{new_pid + 1}",
                start_time=math.inf,
                epochs=[dict(end_size=N0, end_time=time)],
            )
            # In ms, the probability of staying in source is p and the
            # probabilty of moving to the new population is 1 - p.
            b.add_pulse(
                source=f"deme{new_pid + 1}",
                dest=f"deme{pid + 1}",
                time=time,
                proportion=1 - event.p,
            )
            num_demes += 1

            # Expand each migration matrix with a row and column of zeros.
            for migration_matrix in mm_list:
                for row in migration_matrix:
                    row.append(0)
                migration_matrix.append([0 for _ in range(num_demes)])

        ##
        # Demographic events that affect the migration matrix

        elif isinstance(event, MigrationRateChange):
            # -eM t x
            mm = migration_matrix_at(time)
            for j in range(len(mm)):
                for k in range(len(mm)):
                    if j != k:
                        mm[j][k] = event.x / (num_demes - 1)

        elif isinstance(event, MigrationMatrixEntryChange):
            # -m i j x
            # -em t i j x
            pid_i = convert_population_id(event.i)
            pid_j = convert_population_id(event.j)
            if pid_i == pid_j:
                raise ValueError("Cannot set diagonal elements in migration matrix")
            mm = migration_matrix_at(time)
            mm[pid_i][pid_j] = event.rate

        elif isinstance(event, MigrationMatrixChange):
            # -ma M11 M12 M12 ... M21 ...
            # -ema t npop M11 M12 M12 ... M21 ...
            if "-ma" in event.option_strings:
                event.npop = num_demes
            if event.npop != num_demes:
                raise ValueError(
                    f"-ema 'npop' ({event.npop}) doesn't match the current "
                    f"number of demes ({num_demes})"
                )
            _ = migration_matrix_at(time)
            mm = mm_list[0] = copy.deepcopy(event.M)
            # Ms ignores matrix entries for demes that were previously joined
            # (-ej option), and users may deliberately put invalid values
            # here (e.g. 'x'). So we explicitly set these rates to zero.
            for j in joined:
                for k in range(num_demes):
                    if j != k:
                        mm[j][k] = 0
                        mm[k][j] = 0
        else:
            assert False, f"unhandled option: {event}"

    # Resolve/remove growth_rate in oldest epochs.
    for deme in b.data["demes"]:
        start_time = deme.get("start_time", math.inf)
        epoch = deme["epochs"][0]
        growth_rate = epoch.pop("growth_rate", 0)
        if growth_rate != 0:
            if math.isinf(start_time):
                raise ValueError(
                    f"{deme['name']}: growth rate for infinite-length epoch is invalid"
                )
            dt = start_time - epoch["end_time"]
            epoch["start_size"] = epoch["end_size"] * math.exp(-dt * growth_rate)
        else:
            epoch["start_size"] = epoch["end_size"]

    migrations = migrations_from_mm_list(
        mm_list, mm_end_times, [deme["name"] for deme in b.data["demes"]]
    )
    # Rescale rates so they don't have units of 4*N0.
    for migration in migrations:
        migration["rate"] /= 4 * N0
    b.data["migrations"] = migrations

    # Reinsert each deme so that ancestors come before descendants.
    b.data["demes"] = demes_sorted_by_ancestry(b.data["demes"])

    graph = b.resolve()
    return graph


def remap_deme_names(graph: demes.Graph, names: Mapping[str, str]) -> demes.Graph:
    assert sorted(names.keys()) == sorted(deme.name for deme in graph.demes)
    graph = copy.deepcopy(graph)
    for deme in graph.demes:
        deme.name = names[deme.name]
        deme.ancestors = [names[ancestor] for ancestor in deme.ancestors]
    for migration in graph.migrations:
        migration.source = names[migration.source]
        migration.dest = names[migration.dest]
    for pulse in graph.pulses:
        pulse.source = names[pulse.source]
        pulse.dest = names[pulse.dest]
    for k, deme in list(graph._deme_map.items()):
        del graph._deme_map[k]
        graph._deme_map[names[k]] = deme
    return graph


def from_ms(
    command: str,
    *,
    N0: float,
    deme_names: List[str] = None,
) -> demes.Graph:
    """
    Convert an ms demographic model into a demes graph.

    `Hudson's ms <https://doi.org/10.1093/bioinformatics/18.2.337>`_
    uses coalescent units for times (:math:`t`),
    population sizes (:math:`x`), and migration rates (:math:`M`).
    These will be converted to more familiar units using the given
    ``N0`` value (:math:`N_0`) according to the following rules:

    .. math::

        \\text{time (in generations)} &= 4 N_0 t

        \\text{deme size (haploid individuals)} &= N_0 x

        \\text{migration rate (per generation)} &= \\frac{M}{4 N_0}

    :param str command: The ms command line.
    :param float N0:
        The reference population size (:math:`N_0`) used to translate
        from coalescent units. For a ``command`` that specifies a
        :math:`\\theta` value with the ``-t theta`` option,
        this can be calculated as :math:`N_0 = \\theta / (4 \\mu L)`,
        where :math:`\\mu` is the per-generation mutation rate and
        :math:`L` is the length of the sequence being simulated.
    :param list[str] deme_names: A list of names to use for the demes.
        If not specified, demes will be named deme1, deme2, etc.

    :return: The demes graph.
    :rtype: demes.Graph
    """
    args, unknown = parse_ms_args(command)
    if len(unknown) > 0:
        # TODO: do something better here? Pass unknown args back to user?
        logger.warning(f"Ignoring unknown args: {unknown}")

    graph = build_graph(args, N0)
    if deme_names is not None:
        if len(set(deme_names)) != len(graph.demes):
            raise ValueError(
                f"graph has {len(graph.demes)} unique demes, "
                f"but deme_names has {len(set(deme_names))}"
            )
        name_map = dict(zip((f"deme{j+1}" for j in range(len(deme_names))), deme_names))
        graph = remap_deme_names(graph, name_map)
    return graph
